@@ -147,6 +147,19 @@ class Expr:
             self._determ_token = _tokenize_deterministic(type(self), *self.operands)
         return self._determ_token
 
+    def __dask_keys__(self):
+        """The keys for this expression
+
+        This is used to determine the keys of the output collection
+        when this expression is computed.
+
+        Returns
+        -------
+        keys: list
+            The keys for this expression
+        """
+        return [(self._name, i) for i in range(self.npartitions)]
+
     @staticmethod
     def _reconstruct(*args):
         typ, *operands, token, cache = args
@@ -224,7 +237,7 @@ class Expr:
         >>> class Add(Expr):
         ...     def _task(self, i):
         ...         return Task(
-        ...            name,
+        ...            self.__dask_keys__()[i],
         ...            operator.add,
         ...            TaskRef((self.left._name, i)),
         ...            TaskRef((self.right._name, i))
@@ -245,15 +258,23 @@ class Expr:
         )
 
     def _layer(self) -> dict:
-        """The graph layer added by this expression
+        """The graph layer added by this expression.
+
+        Simple expressions that apply one task per partition can choose to only
+        implement `Expr._task` instead.
 
         Examples
         --------
         >>> class Add(Expr):
         ...     def _layer(self):
         ...         return {
-        ...             (self._name, i): (operator.add, (self.left._name, i), (self.right._name, i))
-        ...             for i in range(self.npartitions)
+        ...            name: Task(
+        ...                name,
+        ...                operator.add,
+        ...                TaskRef((self.left._name, i)),
+        ...                TaskRef((self.right._name, i))
+        ...            )
+        ...            for i, name in enumerate(self.__dask_keys__())
         ...         }
 
         Returns
@@ -533,7 +554,17 @@ class Expr:
         return {}
 
     def __dask_graph__(self):
-        """Traverse expression tree, collect layers"""
+        """Traverse expression tree, collect layers
+
+        Subclasses generally do not want to override this method unless custom
+        logic is required to treat (e.g. ignore) specific operands during graph
+        generation.
+
+        See also
+        --------
+        Expr._layer
+        Expr._task
+        """
         stack = [self]
         seen = set()
         layers = []
@@ -1076,31 +1107,24 @@ class _HLGExprSequence(Expr):
         from dask.highlevelgraph import HighLevelGraph
 
         groups = toolz.groupby(
-            lambda x: (
-                x.low_level_optimizer if isinstance(x, HLGExpr) else None,
-                x.postcompute,
-            ),
+            lambda x: x.low_level_optimizer if isinstance(x, HLGExpr) else None,
             self.operands,
         )
         exprs = []
         changed = False
-        for (optimizer, postcompute), group in groups.items():
+        for optimizer, group in groups.items():
             if len(group) > 1:
-                changed = True
-                graphs = []
-                for expr in group:
-                    graphs.append(expr.hlg)
+                graphs = [expr.hlg for expr in group]
 
+                changed = True
                 dsk = HighLevelGraph.merge(*graphs)
-                keys = [v.__dask_keys__() for v in group]
-                exprs.append(
-                    _HLGExprGroup(
-                        dsk=dsk,
-                        low_level_optimizer=optimizer,
-                        output_keys=keys,
-                        postcompute=postcompute,
-                    )
+                hlg_group = _HLGExprGroup(
+                    dsk=dsk,
+                    low_level_optimizer=optimizer,
+                    output_keys=[v.__dask_keys__() for v in group],
+                    postcompute=[g.postcompute for g in group],
                 )
+                exprs.append(hlg_group)
             else:
                 exprs.append(group[0])
         if not changed:
@@ -1302,15 +1326,22 @@ class HLGFinalizeCompute(HLGExpr):
         layers = expr.dsk.layers.copy()
         deps = expr.dsk.dependencies.copy()
         keys = expr.__dask_keys__()
-        func, extra_args = expr.postcompute
-        t = Task(self._name, func, _convert_dask_keys(keys), *extra_args)
+        if isinstance(expr.postcompute, list):
+            postcomputes = expr.postcompute
+        else:
+            postcomputes = [expr.postcompute]
+        tasks = [
+            Task(self._name, func, _convert_dask_keys(keys), *extra_args)
+            for func, extra_args in postcomputes
+        ]
         from dask.highlevelgraph import HighLevelGraph, MaterializedLayer
 
-        layers[t.key] = MaterializedLayer({t.key: t})
         leafs = set(deps)
         for val in deps.values():
             leafs -= val
-        deps[t.key] = leafs
+        for t in tasks:
+            layers[t.key] = MaterializedLayer({t.key: t})
+            deps[t.key] = leafs
         return HighLevelGraph(layers, dependencies=deps)
 
     def __dask_keys__(self):
@@ -1324,6 +1355,7 @@ class ProhibitReuse(Expr):
     """
 
     _parameters = ["expr"]
+    _ALLOWED_TYPES = [HLGExpr, LLGExpr, HLGFinalizeCompute, _HLGExprSequence]
 
     def __dask_keys__(self):
         return self._modify_keys(self.expr.__dask_keys__())
@@ -1345,18 +1377,41 @@ class ProhibitReuse(Expr):
             k = str(k)
         return f"{k}-{self._suffix}"
 
+    def _simplify_down(self):
+        # FIXME: Shuffling cannot be rewritten since the barrier key is
+        # hardcoded. Skipping this here should do the trick most of the time
+        if not isinstance(
+            self.expr,
+            tuple(self._ALLOWED_TYPES),
+        ):
+            return self.expr
+
     def __dask_graph__(self):
+        try:
+            from distributed.shuffle._core import P2PBarrierTask
+        except ModuleNotFoundError:
+            P2PBarrierTask = type(None)
         dsk = convert_legacy_graph(self.expr.__dask_graph__())
 
         subs = {old_key: self._modify_keys(old_key) for old_key in dsk}
-        dsk2 = {
-            new_key: Task(
+        dsk2 = {}
+        for old_key, new_key in subs.items():
+            t = dsk[old_key]
+            if isinstance(t, P2PBarrierTask):
+                warnings.warn(
+                    "Cannot block reusing for graphs including a "
+                    "P2PBarrierTask. This may cause unexpected results. "
+                    "This typically happens when converting a dask "
+                    "DataFrame to delayed objects.",
+                    UserWarning,
+                )
+                return dsk
+            dsk2[new_key] = Task(
                 new_key,
                 ProhibitReuse._identity,
-                dsk.pop(old_key).substitute(subs),
+                t.substitute(subs),
             )
-            for old_key, new_key in subs.items()
-        }
+
         dsk2.update(dsk)
         return dsk2
 
